'Copyright (C) 2009 Raymond Allen.  

'Notes:

CON
  _CLKMODE      = XTAL1 + PLL8X                       
  _XINFREQ      = 10_000_000

'Settings for RS232 host communications:
  RxPin = 31      'Pins 31&30 are also used for programming so they are easiest to use
  TxPin = 30      'But, you could also use pins 18, 19, 20 or 21 by changing these values.
  BaudRate = 115200  'Using this baud rate for serial communications

  ACK = $06       'send when command acknowledged
  NAK = $15      'send when command not understood

  CommandPrefix="~"  

  InputBufferSize=255  'input buffer size in bytes

  'Zero offsets for MEMS 3-axis accelerometer (just in case their off)
  'MEMSXOFF=128
  'MEMSYOFF=128
  'MEMSZOFF=128

OBJ
  lcd : "Proppad_2.8driver"  'Controls the LCD
  sd  : "PSM_SdCardDriver"   'Controls the SD card 
  i2c : "PSM_i2cDriver"  'for AD7999 (used to read data from the 3-axis accelerometer, and extra ADC input)
                         'and AD5337 (used to set LCD brightness, and extra DAC output)
  tone: "PSM_ToneGenerator"   'Generates sine wave tones out the headphone jack
  ser : "FullDuplexSerialExtended"  'using extended version of serial driver for larger buffer 
  num : "numbers"  'Used to turn integers into formatted strings
      

VAR

  'SD card control variables
  long foregroundcolor,backgroundcolor
  byte InputBuffer[InputBufferSize]
  long lineBuffer[320*3/4]' buffer for line of an image  'needs to be long aligned
  byte tbuf[20]  'string buffer for SD


pub Main               'Startup Routine
  'set default colors
  backgroundcolor:=lcd.RGB(0,0,50) 'dark blue
  foregroundcolor:=lcd.RGB(255,255,0) 'yellow
  'initialize the LCD
  lcd.start(0)
  'turn on the LCD's LED backlight
  i2c.SetDAC(0,255)
  'Start the serial driver
  ser.start(RxPin,TxPin,0,BaudRate)
  'Start the SD card driver
  sd.start
  'init numbers
  num.init
  'Begin accepting and processing serial commands
  CommandLoop


      
Pub GetB   'get a byte of command data
  return ser.rx

PUB GetW  'get a word of command data
  return  ser.rx+ser.rx<<8

PUB GetL  'get a Long of command data
  return  ser.rx+ser.rx<<8+ser.rx<<16+ser.rx<<24





PUB CommandLoop| i   ' start command receive&process cycle 
    'Now, start command receive&process cycle
  repeat
    'wait for command code prefix
    repeat
      i:=GetB
    until i==CommandPrefix  'wait for the command code prefix, $C5

    'get command group code
    case GetB
      "G": 'General commands group
        ProcessGeneralCommand
      "T":  'Text commands group
        ProcessTextCommand
      "D": 'Drawing commands group
        ProcessDrawingCommand
      "S":  'SD card commands Group
        ProcessSdCommand
      "A":  'Ancillary commands Group
        ProcessAncillaryCommand

PUB ProcessAncillaryCommand |i,f,d,a          'A prefix
  'Get specific command
  case GetB
    "T": 'generate audio tone
      tone.ToneGen(GetW,GetW,GetB)
      ser.tx(ACK)
    "A": 'Get ADC value
      i:=i2c.GetADC(GetB)
      ser.tx(i)
      ser.tx(i>>8)
      ser.tx(ACK)
    "D":  'Set DAC value
      i2c.SetDAC(GetB,GetW)
      ser.tx(ACK)
    "S":  'Set Free pin state
      i:=GetB
      DIRA[i]:=GetB
      OUTA[i]:=GetB
      ser.tx(ACK)
    "G":  'Get Free pin state
      ser.tx(ACK)
      ser.tx(INA[23..16])
      

PUB ProcessGeneralCommand      'G prefix
  'Get specific command
  case GetB
    "E": 'erase screen with background color
      lcd.Cls(backgroundcolor)
      ser.tx(ACK)
    "B": 'Set Background color
      'lcd.SetBackgroundColor(lcd.RGB(ser.rx,ser.rx,ser.rx))
      backgroundcolor:=lcd.RGB(GetB,GetB,GetB)
      ser.tx(ACK)
    "F": 'Set Foreground color
      'lcd.SetForegroundColor(lcd.RGB(ser.rx,ser.rx,ser.rx))
      foregroundcolor:=lcd.RGB(GetB,GetB,GetB)
      ser.tx(ACK)
    "R": 'Read pixel (set and return as foreground color)
      foregroundcolor:=lcd.GetPixel(GetW,GetW)
      'send back color in 3 bytes
      ser.tx(ACK)
      ser.tx(foregroundcolor>>16) 'R
      ser.tx(foregroundcolor>>8)  'G
      ser.tx(foregroundcolor)     'B
    "I":  'Device info
      ser.str(@DeviceInfoString)
      ser.tx(0) 'end of string
      ser.tx(ACK)
      if ser.rx==1
        ShowDeviceInfo

PUB ShowDeviceInfo
  LCD.CLS(lcd.RGB(0,0,100))
  LCD.DrawString(0,20,50,String("Propeller System"),lcd.RGB(255,255,0),lcd.RGB(0,0,100))
  LCD.DrawString(0,30,82,String("Module Ver. 1a"),lcd.RGB(255,255,0),lcd.RGB(0,0,100))
  LCD.DrawString(1,40,200,String("(C) Raymond Allen 2009"),lcd.RGB(255,255,0),lcd.RGB(0,0,100))          

DAT

DeviceInfoString byte
byte "Propeller System Module Ver.1a",0
      
PUB ProcessTextCommand |i,j,k,x,y,nx,ny       'T prefix
  'Get specific command
  case GetB
    "C":  'draw a character
      lcd.DrawCharacter(GetB,GetW,GetW,GetB,foregroundcolor,backgroundcolor)
      '             (options,  X ,  Y ,Char,  ...          ,      ....     )
      ser.tx(ACK)
    "S":  'Draw a string of characters
      i:=GetB  'options
      x:=GetW  'X
      y:=GetW  'Y
      j:=0
      repeat 
        k:=GetB 
        InputBuffer[j++]:=k
        if j==InputBufferSize  'catch buffer overflow
          InputBuffer[--j]:=0
          quit
      while k>0              
      lcd.DrawString(i,x,y,@InputBuffer,foregroundcolor,backgroundcolor)
      ser.tx(ACK)
            
PUB ProcessDrawingCommand |i,j,k,x,y,nx,ny       'D prefix
  'Get specific command        
  case GetB
    "S": 'set a pixel
      lcd.SetPixel(GetW,GetW,foregroundcolor)
      ser.tx(ACK)
      
    "L": 'draw a line
      lcd.DrawLine(GetW,GetW,GetW,GetW,foregroundcolor)
      ser.tx(ACK)

    "P": 'draw a poly-line  (up to 255 vertices)
      i:=getB
      x:=GetW
      y:=GetW
      repeat i-1
        nx:=GetW
        ny:=GetW
        lcd.DrawLine(x,y,nx,ny,foregroundcolor)
        x:=nx
        y:=ny   
      ser.tx(ACK)
      
    "R":  'Draw a rectangle with options to fill solid or with background color
      'options byte: bit0=Fill Center, bit1=Fill with Background color
      j:=GetB 'get options byte
      case j
        0: 'just the outline of the rectangle in foreground color
          lcd.DrawRectangle(GetW,GetW,GetW,GetW,foregroundcolor)
        1:
          lcd.DrawSolidRectangle(GetW,GetW,GetW,GetW,foregroundcolor)
        2,3:
          lcd.DrawFilledRectangle(GetW,GetW,GetW,GetW,foregroundcolor,backgroundcolor)
      ser.tx(ACK)  
                             
    "T":  'Draw a triangle with options to fill solid or with background color
      'options byte: bit0=Fill Center, bit1=Fill with Background color
      j:=GetB 'get options byte
      case j
        0: 'just the outline of the rectangle in foreground color
          lcd.DrawTriangle(GetW,GetW,GetW,GetW,GetW,GetW,foregroundcolor)
        1:
          lcd.DrawSolidTriangle(GetW,GetW,GetW,GetW,GetW,GetW,foregroundcolor)
        2,3:
          lcd.DrawFilledTriangle(GetW,GetW,GetW,GetW,GetW,GetW,foregroundcolor,backgroundcolor)
      ser.tx(ACK)
   
    "C":  'Draw a circle with options to fill solid or with background color
      'options byte: bit0=Fill Center, bit1=Fill with Background color
      j:=GetB 'get options byte
      case j
        0: 'just the outline of the circle in foreground color
          lcd.DrawCircle(GetW,GetW,GetW,foregroundcolor)
        1:
          lcd.DrawSolidCircle(GetW,GetW,GetW,foregroundcolor)
        2,3:
          lcd.DrawFilledCircle(GetW,GetW,GetW,foregroundcolor,backgroundcolor)
      ser.tx(ACK)            
      
    "I": 'Display image. Image comes as raw 16 or 24-bit data
      x:=GetW
      y:=GetW
      nx:=GetW
      ny:=GetW
      k:=GetB 'mode 24 or 16 bits per pixel

      'lcd.DrawString(4,5,3,num.decf(x,3),foregroundcolor,backgroundcolor)
      'lcd.DrawString(4,5,4,num.decf(y,3),foregroundcolor,backgroundcolor)
      'lcd.DrawString(4,5,5,num.decf(nx,3),foregroundcolor,backgroundcolor)
      'lcd.DrawString(4,5,6,num.decf(ny,3),foregroundcolor,backgroundcolor)
      ' waitcnt(cnt+clkfreq*3)
        
      lcd.SetPosWindow(x,y,nx,ny)  'set starting position and window for image
      'lcd.SetPosition(x,y)  'set starting positionfor image
      case k
        16:
          lcd.WriteReg($03,$5018)  'switch to 2-byte per pixel mode  , and set scan direction

          '         TRI DFM 0 BGR 0 0 HWN 0 0 0 I/D1 I/D0 AM 0 0 0
          '               1     1                        1  1
          ' $5018 = 101000000011000
          
          lcd.WriteCmd($22)  'switch to gram register
          
          'read 16-bit data and send directly to Oled
          repeat nx*ny
            lcd.writeData(GetW)
          ser.tx(ACK)
          'lcd.WriteReg($03,$D038)  'switch back to 3-byte per pixel mode
          lcd.ResetWindow 'reset window and mode
        24:
          lcd.WriteReg($03,$D028)  'set scan direction  danne  $D028
          lcd.WriteCmd($22)  'switch to gram register
          'read 16-bit data and send directly to Oled
          repeat nx*ny
            lcd.writeColor(lcd.rgb(GetB,GetB,GetB))
          ser.tx(ACK)
          'lcd.WriteReg($03,$D038)  'switch back to 3-byte per pixel mode
          lcd.ResetWindow 'reset window                  

PUB ProcessSdCommand|bMountFailure,y,x,i,j  'process SD card commands
  case GetB
    "D":  'View SD Card directory:  transmit directory text as null terminated string
        if not SD.bCardInserted
          ser.str(string("No SD card inserted."))  'no SD card present!
          ser.tx(0)
          return
        bMountFailure:=\SD.mountSD
        if (bMountFailure)
          bMountFailure:=\SD.mountSD  'try again to mount SD card
        if (bMountFailure)                
          ser.str(string("SD card failed to mount."))  'something wrong with SD
          ser.tx(0)
        else
          ser.str(string(" Directory of SD Card: ", 13,10))
          ser.str(string("===================================", 13,10))
          ser.str(string("Filename        Size in Bytes", 13,10))
          ser.str(string("------------    -------------", 13,10))
         
          y:=0 'init total size 
          sd.opendir
          repeat while 0 == sd.nextfile(@tbuf)
            repeat 12-strsize(@tbuf)
             ser.tx(" ")
            ser.str(@tbuf)
            ser.str(string("   "))
            x:=sd.fsize
            y+=x
            ser.str(num.tostr(x,num#DSDEC14))
            ser.tx(13)
            ser.tx(10)
          
          'Show total bytes  
          ser.str(string(13,10,13,10,"Total: "))
          ser.str(num.tostr(y,num#DSDEC14))   
          ser.str(string(13,10,13,10,"bytes"))
          ser.tx(0) 'end of string
          'Note this function returns string instead of ACK or NAK
    "B":  'Show bitmap image
          'get filename
          x:=GetW  'X
          y:=GetW  'Y
          i:=0
          j:=0  'flag to indicate that a "." character recieved
          repeat
            tbuf[i]:=GetB
            if (tbuf[i]==".")
              j:=i
          until tbuf[i++]==0
          if j==0  '.bmp extension not given, so adding it
            bytemove(@tbuf[i-1],string(".bmp"),5)
          if (bMountFailure)
            bMountFailure:=\SD.mountSD  'try again to mount SD card
          if (bMountFailure)                
            ser.tx(NAK)  'something wrong with SD
          else
            ser.tx(ACK) 'OK  
            if BMP(x,y,@tbuf[0])==0 'show bitmap
              


    "P":  'Printscreen (copy screen to BMP file on SD card)
      'Save screen as 24-bit Windows Bitmap to sd card
      'get filename
      i:=0
      j:=0  'flag to indicate that a "." character recieved
      repeat
        tbuf[i]:=GetB
        if (tbuf[i]==".")
          j:=i
      until tbuf[i++]==0
      if j==0  'extension not given, so adding it
        bytemove(@tbuf[i-1],string(".bmp"),5)
        
      if (bMountFailure)
        bMountFailure:=\SD.mountSD  'try again to mount SD card
      if (bMountFailure)                
        ser.tx(NAK)  'something wrong with SD
      else
        if sd.popen(@tbuf,"w")<>0
          ser.tx(NAK)
        else
          ser.tx(ACK) 'OK   'give ack now because about to take some time and write file
          PrintScreen

    "F":  'send file to PC from SD  
      'get filename                      
      i:=0
      j:=0  'flag to indicate that a "." character recieved
      repeat
        tbuf[i]:=GetB
        if (tbuf[i]==".")
          j:=i
      until tbuf[i++]==0

      'send file size if can open file or 0
      if (bMountFailure)
        bMountFailure:=\SD.mountSD  'try again to mount SD card
      if (bMountFailure)                
        ser.tx(NAK)  'something wrong with SD
        return

      if sd.popen(@tbuf,"r")<>0
        ser.tx(NAK)
        return

      ser.tx(ACK)
      'send file size
      i:=sd.fsize
      ser.tx(i)
      ser.tx(i>>8)
      ser.tx(i>>16)
      ser.tx(i>>24)

      x:=200 'output chunk size
      y:=i
      'send file
      repeat until y==0
        if y=>x
          j:=x
        else
          j:=y
        repeat j
          ser.tx(sd.pgetc)
        y-=j
        if ser.rx<>ACK
          return  'error
          
      ser.tx(ACK)
      
      

    "S":  'receive file to SD from PC
      'get filename                      
      i:=0
      j:=0  'flag to indicate that a "." character recieved
      repeat
        tbuf[i]:=GetB
        if (tbuf[i]==".")
          j:=i
      until tbuf[i++]==0
      'get filesize
      i:=GetL
      if (bMountFailure)
        bMountFailure:=\SD.mountSD  'try again to mount SD card
      if (bMountFailure)                
        ser.tx(NAK)  'something wrong with SD
        return  
      if (sd.popen(@tbuf,"w"))==0
        ser.tx(ACK)
      else
        ser.tx(NAK)  'couldn't open file
        return

      j:=GetB     
      repeat while j>0
        x:=0  
        repeat j
          InputBuffer[x++]:=ser.rx
        sd.pwrite(@InputBuffer,j)
        ser.tx(ACK)
        i-=j
        j:=GetB
      if (i==0)
        ser.tx(ACK)
      else
        ser.tx(NAK)
      sd.pclose

    "E":  'Erase file on SD card                               
      'get filename
      i:=0
      repeat
        tbuf[i]:=ser.rx
      until tbuf[i++]==0
      if (bMountFailure)
        bMountFailure:=\SD.mountSD  'try again to mount SD card
      if (bMountFailure)                
        ser.tx(NAK)  'something wrong with SD
      else
        if sd.popen(@tbuf,"d")<>0
          ser.tx(NAK)
        else
          ser.tx(ACK) 'OK   

                
    "X":  'Execute a binary program from SD card                
        'get filename
        i:=0
        j:=0  'flag to indicate that a "." character recieved
        repeat           
          tbuf[i]:=ser.rx
          if (tbuf[i]==".")
            j:=i
        until tbuf[i++]==0
        if j==0  'extension not given, so adding it
          bytemove(@tbuf[i-1],string(".bin"),5)
          
        if (bMountFailure)
          bMountFailure:=\SD.mountSD  'try again to mount SD card
        if (bMountFailure)                
          ser.tx(NAK)  'something wrong with SD
        else                    
          ser.tx(ACK) 'OK   
          i:=SD.bootFile(@tbuf) 'run program

    "W":  'Open file on SD card for writing                               
        'get filename
        i:=0
        repeat
          tbuf[i]:=ser.rx
        until tbuf[i++]==0
        if (bMountFailure)
          bMountFailure:=\SD.mountSD  'try again to mount SD card
        if (bMountFailure)                
          ser.tx(NAK)  'something wrong with SD
        else
          if sd.popen(@tbuf,"w")<>0
            ser.tx(NAK)
          else
            ser.tx(ACK) 'OK                       

    "w":  'Write bytes to SD card file   (up to 255) 
      'file must have already been opened with @W command
      i:=ser.rx  'get # bytes to write
      repeat i
        j:=ser.rx
        sd.pputc(j)
      ser.tx(ACK)

    "R":  'Open file on SD card for reading                                
      'get filename
      i:=0
      repeat
        tbuf[i]:=ser.rx
      until tbuf[i++]==0
      if (bMountFailure)
        bMountFailure:=\SD.mountSD  'try again to mount SD card
      if (bMountFailure)                
        ser.tx(NAK)  'something wrong with SD
      else
        if sd.popen(@tbuf,"r")<0
          ser.tx(NAK)
        else
          ser.tx(ACK) 'OK        

    "r":  'Read bytes from SD card file   (up to 255)
      'file must have already been opened with @R command
      i:=ser.rx  'get # bytes to read
      j:=sd.pread(@InputBuffer,i)
      'Send # bytes actually read
      if j<0
        ser.tx(0)
      else
        ser.tx(j)
      'Transmit bytes
      repeat x from 0 to j-1
        y:=InputBuffer[x]
        ser.tx(y)
      ser.tx(ACK)

    "C":  'close open file on SD card
      sd.pclose
      ser.tx(ACK)       



PUB PrintScreen|x,y,c   'saves display to windows bitmap on SD card file already opened
  'note that this routine is not optimized and takes about 10 seconds to complete!
  'First, write the header 
  biWidth:=320
  biHeight:=240
  bfSize:=biWidth*biHeight*3+54  
  sd.pwrite(@bfType,2)
  sd.pwrite(@bfSize,4)   
  sd.pwrite(@bfReserved1,2)   
  sd.pwrite(@bfReserved2,2)   
  sd.pwrite(@bfOffBits,4)   
  sd.pwrite(@biSize,4)   
  sd.pwrite(@biWidth,4)   
  sd.pwrite(@biHeight,4)   
  sd.pwrite(@biPlanes,2)   
  sd.pwrite(@biBitCount,2)   
  sd.pwrite(@biCompression,4)   
  sd.pwrite(@biSizeImage,4)   
  sd.pwrite(@biXPelsPerMeter,4)   
  sd.pwrite(@biYPelsPerMeter,4)
  sd.pwrite(@biClrUsed,4)   
  sd.pwrite(@biClrImportant,4)
  
  'next, spit out 24-bit color pixels (in reverse order) 
  repeat y from biHeight-1 to 0 
      'vertical line loop
      repeat x from 0 to biWidth-1   'biWidth*3 should be multiple of 4 to avoid padding
        'horizontal loop
        c:=lcd.GetPixel(x,y)  'real pixel read
        sd.pputc(c)
        sd.pputc(c>>8)  
        sd.pputc(c>>16)  

  sd.pclose 


PUB BMP(X,Y,sFilename)|i,j,r,g,b,c,d
'Draw a bitmap file onto the screen
    'note that bitmaps are stored from bottom line to top line!
    i:=\sd.popen(sFilename,"r") 'open image for reading
    if i<>0
      return i   
    sd.pread(@bfType,2) 'read bmp header
    sd.pread(@bfSize,4) 'read bmp header  
    sd.pread(@bfReserved1,4) 'read bmp header  
    sd.pread(@bfOffBits,16) 'read bmp header
    sd.pread(@biPlanes,4)
    sd.pread(@biCompression,24)

    if (biHeight<1) or (biWidth<1)
      return -1  'some kind of problem
      
    case biBitCount
      24:  'true color (3 bytes per pixel)
        lcd.PrepBmp24(x,y,biWidth,biHeight)
        repeat j from biHeight+y-1 to y
          'process image row by row
          sd.pRead(@lineBuffer,biWidth*3) 
          lcd.Write_BmpLine(biBitCount,biWidth,@lineBuffer)
          i:=biWidth//4
          repeat while i>0  'note that each line is padded to have an even # of longs  
            sd.pgetc
            i--

      8:  '256-Palette mode
        'first need to read in palette  (rgbquad (4 bytes) * # pallete entries)
        'calculate actual bytes in palette
        i:=bfOffBits-54
        'read in palette
        sd.pread(@BmpPalette,i)
        'now paint picture
        lcd.PrepBmp24(x,y,biWidth,biHeight)
        repeat j from biHeight+y-1 to y  
          repeat i from x to biWidth+x-1
            c:=sd.pgetc  'read in next pixel
            c:=BmpPalette[c]  'get RGBQuad
            b:=c.byte[0]  
            g:=c.byte[1]  
            r:=c.byte[2]
            lcd.writeColor(lcd.rgb(b,g,r))   'reversing red&blue to be same as 24-bit mode...
          i:=biWidth//4
          repeat while i>0  'note that each line is padded to have an even # of longs  
            sd.pgetc
            i-- 
    
      4:  '16-Palette mode
        'first need to read in palette  (rgbquad (4 bytes) * # pallete entries)
        'calculate actual bytes in palette
        i:=bfOffBits-54
        'read in palette
        sd.pread(@BmpPalette,i)
        'now paint picture
        lcd.PrepBmp24(x,y,biWidth,biHeight) 
        repeat j from biHeight+y-1 to y  
          repeat i from x to biWidth+x-1 step 2
            c:=sd.pgetc  'read in next pixel
            d:=BmpPalette[c>>4]  'get RGBQuad of first pixel
            b:=d.byte[0]  
            g:=d.byte[1]  
            r:=d.byte[2]
            lcd.writeColor(lcd.rgb(b,g,r))   'reversing red&blue to be same as 24-bit mode...
            d:=BmpPalette[c&$F]  'get RGBQuad of second pixel
            b:=d.byte[0]  
            g:=d.byte[1]  
            r:=d.byte[2]  
            lcd.writeColor(lcd.rgb(b,g,r))   'reversing red&blue to be same as 24-bit mode...
          i:=biWidth//4
          repeat while i>0  'note that each line is padded to have an even # of longs  
            sd.pgetc
            i--

    lcd.ResetWindow 'reset window and mode   
    sd.pclose
    return 0

DAT     'Data for reading bitmap files

BMPHeader  byte 'Mostly using info from here:  http://www.fortunecity.com/skyscraper/windows/364/bmpffrmt.html
bfType byte "B","M"  ' 19778
bfSize long 0
bfReserved1 word 0
bfReserved2 word 0
bfOffBits long 54
biSize long 40
biWidth long 0
biHeight long 10
biPlanes word 1
biBitCount word 24
biCompression long 0
biSizeImage long 0
biXPelsPerMeter long 0
biYPelsPerMeter long 0
biClrUsed long 0
biClrImportant long 0

RGB byte  'used for writing 24-bit color data
RGB_b byte 0
RGB_g byte 0
RGB_r byte 0

BmpPalette long 0[256]      'container for bmp palette entries